{
  "type": "record",
  "name": "EditableMLPrimaryKeyProperties",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "doc": "Properties associated with a MLPrimaryKey editable from the UI",
  "fields": [
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Documentation of the MLPrimaryKey",
      "default": null
    }
  ],
  "Aspect": {
    "name": "editableMlPrimaryKeyProperties"
  }
}